using System;

class TreeNode
{
	public TreeNode(int value)
	{
		this.Value = value;
	}
	
	public int Value { get; set; }
	
	public TreeNode Parent { get; set; }
	public TreeNode Left { get; set; }
	public TreeNode Right { get; set; }
}

class BinaryTree
{
	public BinaryTree()
	{
	
	}
	
	public BinaryTree(int[] values)
	{
		foreach(int value in values)
		{
			this.Insert(value);
		}
	}
	
	public void Insert(int value)
	{
		if(Root == null)
		{
			Root = new TreeNode(value);
		}
		else
		{
			TreeNode current = Root;
			
			while(true)
			{
				if(current.Value > value)
				{
					if(current.Left != null)
					{
						current = current.Left;
					}
					else
					{
						current.Left = new TreeNode(value);
						current.Left.Parent = current;
					}
				}
				else if(current.Value < value)
				{
					if(current.Right != null)
					{
						current = current.Right;
					}
					else
					{
						current.Right = new TreeNode(value);
						current.Right.Parent = current;
					}
				}
				else
					break;
			}
		}
	}
	
	public void PreOrder(Action<int> action)
	{
		_preOrder(Root, action);
	}
	
	private void _preOrder(TreeNode root, Action<int> action)
	{
		if(root == null) { return; }
		
		action(root.Value);
		_preOrder(root.Left, action);
		_preOrder(root.Right, action);
	}
	
	public void InOrder(Action<int> action)
	{
		_inOrder(Root, action);
	}

	private void _inOrder(TreeNode root, Action<int> action)
	{
		if(root == null) { return; }
		
		_inOrder(root.Left, action);
		action(root.Value);
		_inOrder(root.Right, action);
	}
	
	public void PostOrder(Action<int> action)
	{
		_postOrder(Root, action);
	}
	
	private void _postOrder(TreeNode root, Action<int> action)
	{
		if(root == null) { return; }
		
		_postOrder(root.Left, action);
		_postOrder(root.Right, action);
		action(root.Value);
	}
	
	public TreeNode Root { get; private set; }
}

static class BinaryTreeHelper
{
	static public void Print(this BinaryTree tree)
	{
		tree.InOrder((item) => Console.WriteLine(item));
	}
}

class Program
{
	static public void Main()
	{
		int[] array = new int[]
		{
			10, 1, 4, 6, 6, 3, 9, 12
		};
		
		BinaryTree tree = new BinaryTree(array);
		tree.Print();
	}
}